"""
parsing.py

Parses X12 270 005010X279A1 segments into a transactional domain model.

The parsing module includes a specific parser for the 270 transaction and loop parsing functions to create new loops
as segments are streamed to the transactional data model.

Loop parsing functions are implemented as set_[description]_loop(context: X12ParserContext, segment_data: Dict).
"""
from enum import Enum

from x12.parsing import match, X12ParserContext
from typing import Optional, Dict


class TransactionLoops(str, Enum):
    """
    The loops used to support the 270 005010X279A1 format.
    """

    HEADER = "header"
    INFORMATION_SOURCE = "loop_2000a"
    INFORMATION_SOURCE_NAME = "loop_2100a"
    INFORMATION_RECEIVER = "loop_2000b"
    INFORMATION_RECEIVER_NAME = "loop_2100b"
    SUBSCRIBER = "loop_2000c"
    SUBSCRIBER_NAME = "loop_2100c"
    SUBSCRIBER_ELIGIBILITY = "loop_2110c"
    DEPENDENT = "loop_2000d"
    DEPENDENT_NAME = "loop_2100d"
    DEPENDENT_ELIGIBILITY = "loop_2110d"
    FOOTER = "footer"


def _get_info_source(context) -> Dict:
    """Returns the current information source"""

    return context.transaction_data[TransactionLoops.INFORMATION_SOURCE][-1]


def _get_info_receiver(context) -> Dict:
    """Returns the current information receiver"""

    info_source = _get_info_source(context)
    return info_source[TransactionLoops.INFORMATION_RECEIVER][-1]


def _get_subscriber(context) -> Dict:
    """Returns the current subscriber"""
    info_receiver = _get_info_receiver(context)
    return info_receiver[TransactionLoops.SUBSCRIBER][-1]


# eligibility 270 loop parsing functions
@match("ST")
def set_header_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set header loop for the 270 transaction set.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """

    context.set_loop_context(
        TransactionLoops.HEADER, context.transaction_data[TransactionLoops.HEADER]
    )


@match("HL", conditions={"hierarchical_level_code": "20"})
def set_information_source_hl_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the Information Source (Payer/Clearinghouse) loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """

    if TransactionLoops.INFORMATION_SOURCE not in context.transaction_data:
        context.transaction_data[TransactionLoops.INFORMATION_SOURCE] = [{}]
    else:
        context.transaction_data[TransactionLoops.INFORMATION_SOURCE].append({})

    info_source = context.transaction_data[TransactionLoops.INFORMATION_SOURCE][-1]
    context.set_loop_context(TransactionLoops.INFORMATION_SOURCE, info_source)


@match("HL", conditions={"hierarchical_level_code": "21"})
def set_information_receiver_hl_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the Information Receiver (Provider) loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """

    info_source = _get_info_source(context)

    if TransactionLoops.INFORMATION_RECEIVER not in info_source:
        info_source[TransactionLoops.INFORMATION_RECEIVER] = [{}]
    else:
        info_source[TransactionLoops.INFORMATION_RECEIVER].append({})

    info_receiver = info_source[TransactionLoops.INFORMATION_RECEIVER][-1]

    context.set_loop_context(TransactionLoops.INFORMATION_RECEIVER, info_receiver)


@match("HL", conditions={"hierarchical_level_code": "22"})
def set_subscriber_hl_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the Subscriber (Member) loop
    Initializes optional TRN segment list.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """

    info_receiver = _get_info_receiver(context)

    if TransactionLoops.SUBSCRIBER not in info_receiver:
        info_receiver[TransactionLoops.SUBSCRIBER] = [{}]
    else:
        info_receiver[TransactionLoops.SUBSCRIBER].append({})

    subscriber = info_receiver[TransactionLoops.SUBSCRIBER][-1]
    subscriber["trn_segment"] = []

    context.set_loop_context(TransactionLoops.SUBSCRIBER, subscriber)


@match("NM1")
def set_entity_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the entity name loop based on the current loop context.
    Initializes optional REF segment list

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """

    new_loop_name: Optional[str] = None

    if context.loop_name == TransactionLoops.INFORMATION_SOURCE:
        new_loop_name = TransactionLoops.INFORMATION_SOURCE_NAME
    elif context.loop_name == TransactionLoops.INFORMATION_RECEIVER:
        new_loop_name = TransactionLoops.INFORMATION_RECEIVER_NAME
    elif context.loop_name == TransactionLoops.SUBSCRIBER:
        new_loop_name = TransactionLoops.SUBSCRIBER_NAME
    elif context.loop_name == TransactionLoops.DEPENDENT:
        new_loop_name = TransactionLoops.DEPENDENT_NAME

    context.loop_container[new_loop_name] = {"ref_segment": []}
    context.set_loop_context(new_loop_name, context.loop_container[new_loop_name])


@match("EQ")
def set_eligibility_inquiry_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the eligibility inquiry loop, 2110C or 2110D, for a subscriber or dependent record.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """

    new_loop_name: Optional[str] = None

    if context.loop_name == TransactionLoops.SUBSCRIBER_NAME:
        new_loop_name = TransactionLoops.SUBSCRIBER_ELIGIBILITY
    elif context.loop_name == TransactionLoops.DEPENDENT_NAME:
        new_loop_name = TransactionLoops.DEPENDENT_ELIGIBILITY

    context.loop_container[new_loop_name] = {"amt_segment": []}
    context.set_loop_context(new_loop_name, context.loop_container[new_loop_name])


@match("HL", conditions={"hierarchical_level_code": "23"})
def set_dependent_hl_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the hierarchy level for a dependent segment.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """
    subscriber = _get_subscriber(context)

    if TransactionLoops.DEPENDENT not in subscriber:
        subscriber[TransactionLoops.DEPENDENT] = [{}]
    else:
        subscriber[TransactionLoops.DEPENDENT].append({})

    dependent = subscriber[TransactionLoops.DEPENDENT][-1]
    dependent["trn_segment"] = []

    context.set_loop_context(TransactionLoops.DEPENDENT, dependent)


@match("SE")
def set_se_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set footer loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment's data
    """

    context.set_loop_context(
        TransactionLoops.FOOTER, context.transaction_data["footer"]
    )



"""
cli.py

The LinuxForHealth X12 Command Line Interface.

Usage:
    pipenv run cli --help

"""

import argparse
import json
from x12.encoding import X12JsonEncoder
from x12.io import X12SegmentReader, X12ModelReader
from typing import List, Dict


CLI_DESCRIPTION = """
The LinuxForHealth X12 CLI parses and validates X12 messages.
Messages are returned in JSON format in either a segment or transactional format.
"""


def _create_arg_parser():
    """
    Creates the Argument Parser for the CLI utility.
    :return: ArgumentParser
    """

    parser = argparse.ArgumentParser(
        prog="LinuxForHealth X12",
        description=CLI_DESCRIPTION,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument(
        "-s", "--segment", help="Returns X12 segments", action="store_true"
    )
    mode_group.add_argument(
        "-m", "--model", help="Returns X12 models", action="store_true"
    )

    parser.add_argument(
        "-x",
        "--exclude",
        help="Exclude fields set to None in model output",
        action="store_true",
    )
    parser.add_argument(
        "-p", "--pretty", help="Pretty print output", action="store_true"
    )

    parser.add_argument("file", help="The path to a ASC X12 file")

    return parser.parse_args()


def _parse_segments(file_path: str) -> List:
    """
    Parses X12 segments from an input file.

    :param file_path: The path to the X12 file.
    :return: List of X12 segments
    """

    with X12SegmentReader(file_path) as r:
        segments = []

        for segment_name, segment in r.segments():
            segment_data = {
                f"{segment_name}{str(i).zfill(2)}": v for i, v in enumerate(segment)
            }
            segments.append(segment_data)
        return segments


def _parse_models(file_path: str, exclude_none: bool = False) -> List:
    """
    Parses a X12 segment model from a X12 input file.

    :param file_path: The path to the X12 file.
    :param exclude_none: Excludes fields set to None from the model output.
    :return: List of X12 models
    """

    with X12ModelReader(file_path) as r:
        models = []
        for m in r.models():
            model_data = m.dict(exclude_unset=exclude_none, exclude_none=exclude_none)
            models.append(model_data)
        return models


def main():
    """
    CLI module entrypoint
    """
    args = _create_arg_parser()

    x12_data = (
        _parse_segments(args.file)
        if args.segment
        else _parse_models(args.file, args.exclude)
    )

    json_opts = {"cls": X12JsonEncoder}

    if args.pretty:
        json_opts["indent"] = 4

    x12_json = json.dumps(x12_data, **json_opts)
    print(x12_json)
